-------------------------------------------------------------------------------
-- FILE interrupt_controller.vhd
--		ENTITY interrupt_controller
--			ARCHITECTURE behaviour
--
-- Simple interrupt controller for the X32, it supports multiple interrupt
-- lines, and a basic priority system.
--
-- Devices may be connected to the interrupt controller through the 'irqs' 
-- signal, each device gets its own 1 bit signal, the device may cause an
-- interrupt by driving the line high. The highest index in the irqs vector
-- gets the highest priority, irqs(0) gets the lowest priority, two devices
-- may share a priority, but the software will no longer be able to detect
-- which of the devices caused the interrupt (which is done through the
-- interrupt priority).
--
-- The interrupt controller checks each device sequentially, this causes an
-- average delay of n/2 clockcycles + the processors response time to an
-- interrupt (n is the number of devices passed to the generic map), when
-- only one device is interrupting. In a worst case scenario (all devices
-- are interrupting), the highest priority interrupt may be up to n times
-- the processors response time, which may be up to about 30 clockcycles.
-- Note that this is still that not much compared to the time it'll take to
-- execute a "usefull" software interrupt handler.
--
-- Author: Sijmen Woutersen
-- Last modified: May 3, 2006
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity interrupt_controller is
	generic (
		-- number of interrupt lines to support
		devices						: in  positive;
		-- size (in bits) of priority & address
		size							: in  positive;
		-- size (in bits) of ram address
		address_bits			: in  positive;
		-- one interrupt may be critical: 
		--  it will not be stopped by the global interrupt
		critical_interrupt: in  natural
	);
	port (
		-- system clock
		clk								: in  std_logic;
		-- system reset
		reset							: in  std_logic;
		
		-- interrupt request lines
		irqs							: in  std_logic_vector(devices-1 downto 0);
		ie								: in  std_logic_vector(devices downto 0);
		
		
		-- raise interrupt signal
		interrupt					: out std_logic;
		-- interrupt acknowledge signal
		interrupt_ack			: in  std_logic;
		
		-- current processor priority (execution level)
		current_priority	: in  std_logic_vector(size-1 downto 0);
		-- new processor priority (execution level)
		priority					: out std_logic_vector(size-1 downto 0);
		-- interrupt handler
		handler_address		: out std_logic_vector(size-1 downto 0);

		-- interrupt controller ram
		ram_in						: in  std_logic_vector(size-1 downto 0);
		ram_out						: out std_logic_vector(size-1 downto 0);
		ram_address				: in  std_logic_vector(address_bits-1 downto 0);
		ram_write					: in  std_logic;
		ram_read					: in  std_logic;
		ram_ready					: out std_logic
	);
end entity;

architecture behaviour of interrupt_controller is
	signal current_device : integer range 0 to devices-1;
	
	signal raise_interrupt	: std_logic;
	signal count : std_logic;
	
	signal irq_reg : std_logic_vector(devices-1 downto 0);
	
	type STATETYPE is (RESET_STATE, CHECKING, INTERRUPTING, WRITE, READY);
	signal state_i, state_o : STATETYPE;

	type RAM is array(devices-1 downto 0) of std_logic_vector(size-1 downto 0);
	signal vector_memory : RAM;
	signal priority_memory : RAM;

	signal interrupt_priority : std_logic_vector(size-1 downto 0);
	signal interrupt_vector : std_logic_vector(size-1 downto 0);


	signal sig_ram_write : std_logic;
begin
	ram_out <= interrupt_vector when ram_address(0) = '0' else interrupt_priority;

	priority <= interrupt_priority;
	handler_address <= interrupt_vector;
	
	raise_interrupt <= irq_reg(current_device) when interrupt_priority > current_priority
		and (ie(devices) = '1' or current_device < critical_interrupt) else '0';
	
	-- state register
	process (clk, reset) begin
		if (clk'event and clk = '1') then
			if (reset = '1') then	
				state_o <= RESET_STATE;
			else
				state_o <= state_i;
			end if;
		end if;
	end process;	
	
	-- irq registers
	process(clk, reset, irqs, interrupt_ack, current_device, ie) begin
		if (clk'event and clk = '1') then
			if (reset = '1') then
				irq_reg <= (others => '0');
			else
				for i in 0 to devices-1 loop
					if (irqs(i) = '1' and ie(i) = '1') then
						-- set on irq
						irq_reg(i) <= '1';
					elsif ((interrupt_ack = '1' or ie(i) = '0') and current_device = i) then
						-- reset on ack or disabled
						irq_reg(i) <= '0';
					end if;
				end loop;
			end if;
		end if;
	end process;
	
	-- current device counter
	process(clk, reset, count, current_device) begin
		if (clk'event and clk = '1') then
			if (reset = '1') then
				current_device <= 0;
			elsif (ram_read = '1' or ram_write = '1') then
				current_device <= conv_integer(ram_address(address_bits-1 downto 1));
			elsif (count = '1') then
				if (current_device = devices-1) then
					current_device <= 0;
				else
					current_device <= current_device + 1;
				end if;
			end if;
		end if;
	end process;
	
	-- ram
	interrupt_vector <= vector_memory(current_device);
	interrupt_priority <= priority_memory(current_device);
	process(clk, reset) begin
		if (clk'event and clk = '1') then
			if (sig_ram_write = '1') then
				if (ram_address(0) = '0') then
					vector_memory(current_device) <= ram_in;
				else
					priority_memory(current_device) <= ram_in;
				end if;
			end if;
		end if;
	end process;

	-- fsm
	process(state_o, raise_interrupt, interrupt_ack, ram_read, ram_write) begin
		case state_o is
			when RESET_STATE =>
				interrupt <= '0';
				count <= '0';
				ram_ready <= '0';
				sig_ram_write <= '0';
				state_i <= CHECKING;
			when CHECKING =>
				interrupt <= '0';
				ram_ready <= '0';
				sig_ram_write <= '0';

				if (ram_read = '1') then
					count <= '-';
					state_i <= READY;
				elsif (ram_write = '1') then
					count <= '-';
					state_i <= WRITE;
				elsif (raise_interrupt = '1') then
					count <= '0';
					state_i <= INTERRUPTING;
				else
					count <= '1';
					state_i <= CHECKING;
				end if;
			when INTERRUPTING =>
				interrupt <= '1';
				count <= '0';
				ram_ready <= '0';
				sig_ram_write <= '0';
				
				-- note, when receiving an ram_read/ram_write signal after
				-- raising the interrupt signal, the core hasn't 'detected' the
				-- interrupt request yet, so it's safe to cancel (as the processor
				-- will now wait for a ram_ready signal).
				if (interrupt_ack = '1') then
					state_i <= CHECKING;
				elsif (ram_read = '1') then
					state_i <= READY;
				elsif (ram_write = '1') then
					state_i <= WRITE;
				else
					state_i <= INTERRUPTING;
				end if;
			when WRITE =>
				interrupt <= '0';
				count <= '0';
				ram_ready <= '0';
				sig_ram_write <= '1';
				state_i <= READY;
			when READY =>
				interrupt <= '0';
				count <= '0';
				sig_ram_write <= '0';
				ram_ready <= '1';
				state_i <= CHECKING;
		end case;
	end process;
	
end architecture;
