
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.types.all;

entity mem_decoder is
	port (
		address_offset 	: in  std_logic_vector(1 downto 0);
		mem_data_size		: in  std_logic_vector(2 downto 0);
		mem_data_signed	: in  std_logic;
		
		mem_data_in1		: in  std_logic_vector(31 downto 0);
		mem_data_in2		: in  std_logic_vector(31 downto 8);
		proc_data_in		: out std_logic_vector(31 downto 0);
		
		proc_data_out		: in  std_logic_vector(31 downto 0);
		mem_data_out1		: out std_logic_vector(31 downto 0);
		mem_data_out2		: out std_logic_vector(31 downto 0);

		bytes_enable1		: out std_logic_vector(3 downto 0);
		bytes_enable2		: out std_logic_vector(3 downto 0);

		overflow		: out std_logic
	);
end entity;

architecture behaviour of mem_decoder is
	signal is_32, is_16, is_8 : std_logic;
	signal overflow_16, overflow_8 : std_logic;
	signal overflow_16_signed, overflow_8_signed : std_logic;
	signal ones, zeroes : std_logic_vector(23 downto 0);
begin
	is_32 <= 
		'1' when mem_data_size = VARTYPE_PTR else
		'1' when mem_data_size = VARTYPE_LNG else
		'1' when mem_data_size = VARTYPE_INT else
		'0';
	is_16 <=
		'1' when mem_data_size = VARTYPE_SHORT else
		'1' when mem_data_size = VARTYPE_INSTRUCTION else
		'0';
	is_8 <=
		'1' when mem_data_size = VARTYPE_CHAR else
		'0';

	ones <= (others => '1');
	zeroes <= (others => '0');

	overflow_16_signed <= '0' when proc_data_out(15) = proc_data_out(16) else '1';
	overflow_16 <= 
		'0' when proc_data_out(31 downto 16) = zeroes(15 downto 0) else
		'0' when proc_data_out(31 downto 16) = ones(15 downto 0) else
		'1';

	overflow_8_signed <= '0' when proc_data_out(7) = proc_data_out(8) else '1';
	overflow_8 <= 
		'0' when proc_data_out(31 downto 8) = zeroes(23 downto 0) else
		'0' when proc_data_out(31 downto 8) = ones(23 downto 0) else
		'1';

	overflow <= (overflow_16 and is_16) or (overflow_8 and is_8) or
		(overflow_16_signed and mem_data_signed and is_16) or
		(overflow_8_signed and mem_data_signed and is_8);
			
	process(is_32, is_16, is_8, address_offset, mem_data_signed, mem_data_in1, mem_data_in2, proc_data_out) begin
		if (is_32 = '1' and address_offset = "00") then
			proc_data_in <= mem_data_in1;
		
			mem_data_out1 <= proc_data_out;
			mem_data_out2 <= (others => '-');
		
			bytes_enable1 <= "0000";
			bytes_enable2 <= "1111";
		elsif (is_32 = '1' and address_offset = "01") then
			proc_data_in(31 downto 8) <= mem_data_in1(23 downto 0);
			proc_data_in(7 downto 0) <= mem_data_in2(31 downto 24);

			mem_data_out1(31 downto 24) <= (others => '-');
			mem_data_out1(23 downto 0) <= proc_data_out(31 downto 8);
			mem_data_out2(31 downto 24) <= proc_data_out(7 downto 0);
			mem_data_out2(23 downto 0) <= (others => '-');

			bytes_enable1 <= "1000";
			bytes_enable2 <= "0111";
		elsif (is_32 = '1' and address_offset = "10") then
			proc_data_in(31 downto 16) <= mem_data_in1(15 downto 0);
			proc_data_in(15 downto 0) <= mem_data_in2(31 downto 16);

			mem_data_out1(31 downto 16) <= (others => '-');
			mem_data_out1(15 downto 0) <= proc_data_out(31 downto 16);
			mem_data_out2(31 downto 16) <= proc_data_out(15 downto 0);
			mem_data_out2(15 downto 0) <= (others => '-');

			bytes_enable1 <= "1100";
			bytes_enable2 <= "0011";
		elsif (is_32 = '1' and address_offset = "11") then
			proc_data_in(31 downto 24) <= mem_data_in1(7 downto 0);
			proc_data_in(23 downto 0) <= mem_data_in2(31 downto 8);

			mem_data_out1(31 downto 8) <= (others => '-');
			mem_data_out1(7 downto 0) <= proc_data_out(31 downto 24);
			mem_data_out2(31 downto 8) <= proc_data_out(23 downto 0);
			mem_data_out2(7 downto 0) <= (others => '-');

			bytes_enable1 <= "1110";
			bytes_enable2 <= "0001";
		elsif (is_16 = '1' and address_offset = "00") then
			proc_data_in(31 downto 16) <= (others => (mem_data_signed and mem_data_in1(31)));
			proc_data_in(15 downto 0) <= mem_data_in1(31 downto 16);

			mem_data_out1(31 downto 16) <= proc_data_out(15 downto 0);
			mem_data_out1(15 downto 0) <= (others => '-');
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "0011";
			bytes_enable2 <= "1111";
		elsif (is_16 = '1' and address_offset = "01") then
			proc_data_in(31 downto 16) <= (others => (mem_data_signed and mem_data_in1(23)));
			proc_data_in(15 downto 0) <= mem_data_in1(23 downto 8);

			mem_data_out1(31 downto 24) <= (others => '-');
			mem_data_out1(23 downto 8) <= proc_data_out(15 downto 0);
			mem_data_out1(7 downto 0) <= (others => '-');
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "1001";
			bytes_enable2 <= "1111";
		elsif (is_16 = '1' and address_offset = "10") then
			proc_data_in(31 downto 16) <= (others => (mem_data_signed and mem_data_in1(15)));
			proc_data_in(15 downto 0) <= mem_data_in1(15 downto 0);

			mem_data_out1(31 downto 16) <= (others => '-');
			mem_data_out1(15 downto 0) <= proc_data_out(15 downto 0);
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "1100";
			bytes_enable2 <= "1111";
		elsif (is_16 = '1' and address_offset = "11") then
			proc_data_in(31 downto 16) <= (others => (mem_data_signed and mem_data_in1(7)));
			proc_data_in(15 downto 8) <= mem_data_in1(7 downto 0);
			proc_data_in(7 downto 0) <= mem_data_in2(31 downto 24);

			mem_data_out1(31 downto 8) <= (others => '-');
			mem_data_out1(7 downto 0) <= proc_data_out(15 downto 8);
			mem_data_out2(31 downto 24) <= proc_data_out(7 downto 0);
			mem_data_out2(23 downto 0) <= (others => '-');

			bytes_enable1 <= "1110";
			bytes_enable2 <= "0111";
		elsif (is_8 = '1' and address_offset = "00") then
			proc_data_in(31 downto 8) <= (others => (mem_data_signed and mem_data_in1(31)));
			proc_data_in(7 downto 0) <= mem_data_in1(31 downto 24);

			mem_data_out1(31 downto 24) <= proc_data_out(7 downto 0);
			mem_data_out1(23 downto 0) <= (others => '-');
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "0111";
			bytes_enable2 <= "1111";
		elsif (is_8 = '1' and address_offset = "01") then
			proc_data_in(31 downto 8) <= (others => (mem_data_signed and mem_data_in1(23)));
			proc_data_in(7 downto 0) <= mem_data_in1(23 downto 16);

			mem_data_out1(31 downto 24) <= (others => '-');
			mem_data_out1(23 downto 16) <= proc_data_out(7 downto 0);
			mem_data_out1(15 downto 0) <= (others => '-');
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "1011";
			bytes_enable2 <= "1111";
		elsif (is_8 = '1' and address_offset = "10") then
			proc_data_in(31 downto 8) <= (others => (mem_data_signed and mem_data_in1(15)));
			proc_data_in(7 downto 0) <= mem_data_in1(15 downto 8);

			mem_data_out1(31 downto 16) <= (others => '-');
			mem_data_out1(15 downto 8) <= proc_data_out(7 downto 0);
			mem_data_out1(7 downto 0) <= (others => '-');
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "1101";
			bytes_enable2 <= "1111";
		elsif (is_8 = '1' and address_offset = "11") then
			proc_data_in(31 downto 8) <= (others => (mem_data_signed and mem_data_in1(7)));
			proc_data_in(7 downto 0) <= mem_data_in1(7 downto 0);

			mem_data_out1(31 downto 8) <= (others => '-');
			mem_data_out1(7 downto 0) <= proc_data_out(7 downto 0);
			mem_data_out2(31 downto 0) <= (others => '-');

			bytes_enable1 <= "1110";
			bytes_enable2 <= "1111";
		else
			proc_data_in <= (others => '-');
			mem_data_out1 <= (others => '-');
			mem_data_out2 <= (others => '-');
			
			bytes_enable1 <= "1111";
			bytes_enable2 <= "1111";
		end if;
	end process;
end architecture;
